@@ -35,7 +35,6 @@
  DEF_ASM(rbp)
  DEF_ASM(rsi)
  DEF_ASM(rdi)
- DEF_ASM(rip)
 #endif
  DEF_ASM(mm0)
  DEF_ASM(mm1)
@@ -92,7 +91,16 @@
  DEF_ASM(fs)
  DEF_ASM(gs)
  DEF_ASM(st)
+ DEF_ASM(rip)
 
+#ifdef TCC_TARGET_X86_64
+ /* The four low parts of sp/bp/si/di that exist only on
+    x86-64 (encoding aliased to ah,ch,dh,dh when not using REX). */
+ DEF_ASM(spl)
+ DEF_ASM(bpl)
+ DEF_ASM(sil)
+ DEF_ASM(dil)
+#endif
  /* generic two operands */
  DEF_BWLX(mov)
 
@@ -185,6 +193,7 @@
  DEF_WLX(btr)
  DEF_WLX(btc)
 
+ DEF_WLX(lar)
  DEF_WLX(lsl)
 
  /* generic FP ops */
